#include "source.h"
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>

static void lf_free(loaded_file_t *lf)
{
    if(lf) {
        free(lf->path);
        buffer_free(lf->buf);
    }
    free(lf);
}

static ssize_t lf_read(void *_cookie, void *buf, off_t offset, size_t size)
{
    loaded_file_t *lf = _cookie;
    if(offset < 0 || offset >= (int)lf->buf->data_size)
        return -1;
    if(offset + size > lf->buf->data_size)
        size = lf->buf->data_size - offset;

    memcpy(buf, lf->buf->mem + offset, size);
    return size;
}

static bool lf_write(void *_cookie, void *data, size_t data_size, off_t offset,
    size_t segment_size)
{
    loaded_file_t *lf = _cookie;
    return buffer_write(lf->buf, data, data_size, offset, segment_size);
}

static size_t lf_size(void *_cookie)
{
    loaded_file_t *lf = _cookie;
    return lf->buf->data_size;
}

static bool lf_save(void *_cookie)
{
    loaded_file_t *lf = _cookie;

    /* If the new file is larger than the original, we can just overwrite it
       without truncating, which is much faster with Fugue */
    int flags = O_WRONLY | O_TRUNC;
    if(lf->buf->data_size >= lf->original_size)
        flags = O_WRONLY;
    int fd = open(lf->path, flags);
    if(!fd) goto fail;

    ssize_t rc = write(fd, lf->buf->mem, lf->buf->data_size);
    if(rc < 0 || (size_t)rc != lf->buf->data_size) goto fail;

    close(fd);
    return true;

fail:
    close(fd);
    return false;
}

static void lf_close(void *_cookie)
{
    loaded_file_t *lf = _cookie;
    lf_free(lf);
}

static source_intf_t lf_intf = {
    .cap = SOURCE_WRITE | SOURCE_EXPAND | SOURCE_SHRINK,
    .read = lf_read,
    .write = lf_write,
    .size = lf_size,
    .save = lf_save,
    .close = lf_close,
};

source_t *source_loaded_file_open(char const *path)
{
    int fd = -1;
    source_t *source = NULL;
    loaded_file_t *lf = NULL;

    lf = calloc(1, sizeof *lf);
    if(!lf) goto fail;

    lf->path = strdup(path);
    if(!lf->path) goto fail;

    fd = open(path, O_RDONLY);
    if(fd < 0) goto fail;

    int size = lseek(fd, 0, SEEK_END);
    if(size < 0) goto fail;

    lf->original_size = size;
    lf->buf = buffer_create(size, 4096);
    if(!lf->buf) goto fail;

    lseek(fd, 0, SEEK_SET);
    ssize_t rc = read(fd, lf->buf->mem, size);
    lf->buf->data_size = size;
    if(rc != size) goto fail;

    source = source_open(&lf_intf, lf, path);
    if(!source) goto fail;

    close(fd);
    return source;

fail:
    free(source);
    close(fd);
    lf_free(lf);
    return NULL;
}
